@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 
-__author__ = 'Scott Burns <scott.s.burns@vanderbilt.edu>'
+__author__ = 'Scott Burns <scott.s.burns@gmail.com>'
 __license__ = 'MIT'
 __copyright__ = '2014, Vanderbilt University'
 
@@ -15,12 +15,12 @@ __author__ = 'Scott Burns'
 __copyright__ = ' Copyright 2014, Vanderbilt University'
 
 
-from requests import post, RequestException
+from requests import RequestException, Session
 import json
 
-
 RedcapError = RequestException
 
+_session = Session()
 
 class RCAPIError(Exception):
     """ Errors corresponding to a misuse of the REDCap API """
@@ -40,7 +40,7 @@ class RCRequest(object):
     biggest consumer.
     """
 
-    def __init__(self, url, payload, qtype):
+    def __init__(self, url, payload, qtype, session=_session):
         """
         Constructor
 
@@ -56,6 +56,8 @@ class RCRequest(object):
         self.url = url
         self.payload = payload
         self.type = qtype
+        self.session = session
+
         if qtype:
             self.validate()
         fmt_key = 'returnFormat' if 'returnFormat' in payload else 'format'
@@ -67,6 +69,8 @@ class RCRequest(object):
         valid_data = {
             'exp_record': (['type', 'format'], 'record',
                 'Exporting record but content is not record'),
+            'del_record': (['format'], 'record',
+                'Deleting record but content is not record'),
             'imp_record': (['type', 'overwriteBehavior', 'data', 'format'],
                 'record', 'Importing record but content is not record'),
             'metadata': (['format'], 'metadata',
@@ -83,8 +87,18 @@ class RCRequest(object):
                 'Exporting arms but content is not arm'),
             'exp_fem': (['format'], 'formEventMapping',
                 'Exporting form-event mappings but content != formEventMapping'),
+            'exp_next_id': ([], 'generateNextRecordName',
+                'Generating next record name but content is not generateNextRecordName'),
+            'exp_proj': (['format'], 'project',
+                'Exporting project info but content is not project'),
             'exp_user': (['format'], 'user',
-                'Exporting users but content is not user')
+                'Exporting users but content is not user'),
+            'exp_survey_participant_list': (['instrument'], 'participantList',
+                'Exporting Survey Participant List but content != participantList'),
+            'exp_report': (['report_id', 'format'], 'report',
+                'Exporting Reports but content is not reports'),
+            'version': (['format'], 'version',
+                'Requesting version but content != version')
         }
         extra, req_content, err_msg = valid_data[self.type]
         required.extend(extra)
@@ -92,7 +106,7 @@ class RCRequest(object):
         pl_keys = set(self.payload.keys())
         # if req is not subset of payload keys, this call is wrong
         if not set(required) <= pl_keys:
-            #what is not in pl_keys?
+            # what is not in pl_keys?
             not_pre = required - pl_keys
             raise RCAPIError("Required keys: %s" % ', '.join(not_pre))
         # Check content, raise with err_msg if not good
@@ -108,7 +122,7 @@ class RCRequest(object):
         Parameters
         ----------
         kwargs :
-            passed to requests.post()
+            passed to requests.Session.post()
 
         Returns
         -------
@@ -116,17 +130,19 @@ class RCRequest(object):
             data object from JSON decoding process if format=='json',
             else return raw string (ie format=='csv'|'xml')
         """
-        r = post(self.url, data=self.payload, **kwargs)
+        response = self.session.post(self.url, data=self.payload, **kwargs)
         # Raise if we need to
-        self.raise_for_status(r)
-        content = self.get_content(r)
-        return content, r.headers
+        self.raise_for_status(response)
+        content = self.get_content(response)
+        return content, response.headers
 
     def get_content(self, r):
         """Abstraction for grabbing content from a returned response"""
         if self.type == 'exp_file':
             # don't use the decoded r.text
             return r.content
+        elif self.type == 'version':
+            return r.content
         else:
             if self.fmt == 'json':
                 content = {}
@@ -162,3 +178,6 @@ class RCRequest(object):
         # specifically 10.5
         if 500 <= r.status_code < 600:
             raise RedcapError(r.content)
+
+        if 400 == r.status_code and self.type == 'exp_record':
+            raise RedcapError(r.content)
